#include "ADC.h"
#include <avr/io.h> 
#include <string.h>
#include <stdint.h>
#include "../hw-layout.h"


// Functions to access the internal ADC

#ifdef ADCA_CTRLA
    // On Xmega targets
    
    // Note: For all functional registers, I used the old syntax because
    // the structures of the ADC are incorrectly defined in WinAVR.

    // get a sample from ADC chip.
    // chanell must be in range 0-15.
    uint16_t getADC(const int channel) {
        ADCA_CH0_MUXCTRL = (channel<<3);    
        // flush the previous result
        ADCA_CTRLA |= 2; // 2=ADC_FLUSH_bm but not defined in WinAVR
        // start one sample
        ADCA_CH0_INTFLAGS=1;
        ADCA_CH0_CTRL |= ADC_CH_START_bm;
        // wait for the result
        while((ADCA_CH0_INTFLAGS & 1)==0);
        return ADCA_CH0RES;
    }

    // initialize the ADC 
    void initADC() {
        ADCA_CTRLA = ADC_ENABLE_bm;
        ADCA_CTRLB = ADC_RESOLUTION_12BIT_gc;
        ADCA_REFCTRL = ADC_REFSEL_INT1V_gc | ADC_BANDGAP_bm;
        ADCA_PRESCALER = ADC_PRESCALER_DIV8_gc; 
        ADCA_CH0_CTRL = ADC_CH_INPUTMODE_SINGLEENDED_gc; 
    }

    // select reference
    // The value can be "INT1V", "INTVCC", "AREFA" or "AREFB"
    // returns 1 on success
    uint8_t setAREF(const char* name) {
        if (strcmp(name,"INT1V")==0)
            ADCA_REFCTRL = ADC_REFSEL_INT1V_gc | ADC_BANDGAP_bm; 
        else if (strcmp(name,"INTVCC")==0)
            ADCA_REFCTRL = ADC_REFSEL_VCC_gc   | ADC_BANDGAP_bm; 
        else if (strcmp(name,"AREFA")==0)
            ADCA_REFCTRL = ADC_REFSEL_AREFA_gc | ADC_BANDGAP_bm; 
        else if (strcmp(name,"AREFB")==0)     
            ADCA_REFCTRL = ADC_REFSEL_AREFB_gc | ADC_BANDGAP_bm;                        
        else 
            return 0;
        // Perform one sample to initialize the ADC on the new reference
        getADC(0);
        return 1;
    }

#else
    // On ATmega targets

    // internal ADC reference configuration
    static uint8_t adcReference=(1<<REFS0); // = VCC 5V

    // get a sample from ADC chip.
    // chanell must be in range 0-7.
    uint16_t getADC(const int channel) {
        ADMUX = adcReference + channel;
        // start one sample
        ADCSRA |= (1<<ADSC);
        // wait for the result
        loop_until_bit_is_clear(ADCSRA, ADSC);
        return ADC;
    }

    // initialize the ADC 
    void initADC() {
        ADCSRA |= (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);
    }

    // select reference
    // On ATmega, the value can be "VCC", "1.1V or "2.56V"
    // On ATxmega, the value can be "INT1V", "INTVCC", "AREFA" or "AREFB"
    // returns 1 on success
    uint8_t setAREF(const char* name) {
        if (strcmp(name,"VCC")==0)
            adcReference=(1<<REFS0);     
        else if (strcmp(name,"1.1V")==0)
            adcReference=(1<<REFS1);
        else if (strcmp(name,"2.56V")==0)
            adcReference=((1<<REFS1) | (1<<REFS0));
        else if (strcmp(name,"EXT")==0)     
            adcReference=0;                        
        else 
            return 0;
        // Perform one sample to initialize the ADC on the new reference
        getADC(0);
        return 1;
    }
#endif

