#include "serialconsole.h"
#include <stdio.h>
#include <avr/io.h>
#include <stdint.h>
#include "../hw-layout.h"
#include <avr/interrupt.h> 


// Files for input and output through serial port.
static int serial_write(char, FILE *);
static int serial_read(FILE *);
static FILE serialPort = FDEV_SETUP_STREAM(serial_write, serial_read, _FDEV_SETUP_RW);

// input ring buffer
static volatile char inputBuffer[SERIAL_INPUT_BUFFER_SIZE];

// index of first character in the buffer
static volatile uint8_t inputStart=0;

// point to the next free position in the input buffer (=inputStart in case of empty buffer)
static volatile uint8_t inputEnd=0;

// The previously received character
static volatile char prevChar=0;

// Check if the serial port has received at least one character
uint8_t receivedChar(void) {
    return inputStart!=inputEnd;
}

// Check if the serial port has received a line break
uint8_t receivedLine(void) {
    int i=inputStart;
    while (i!=inputEnd) {
        if (inputBuffer[i]=='\n') {
            return 1;
        }
        if (++i>=SERIAL_INPUT_BUFFER_SIZE) {
            i=0;
        }
    }
    return 0;
}

// Define aliases for ATmega controllers with multiple serial ports
// so they can be used with the same source code as controllers with only a single serial port.
#ifdef TXEN0   // Atmega with multiple serial ports
    #ifdef USE_SERIAL2  // configured to use the second serial port
        #define UCSRA UCSR1A
        #define UCSRB UCSR1B
        #define UCSRC UCSR1C
        #define UBRRH UBRR1H
        #define UBRRL UBRR1L
        #define UDRE  UDRE1
        #define UDR   UDR1
        #define RXC   RXC1
        #define RXEN  RXEN1
        #define TXEN  TXEN1
        #define UCSZ0 UCSZ10
        #define UCSZ1 UCSZ11
        #define U2X   U2X1
        #define RXCIE RXCIE1
        #ifdef USART_RX_vect
            #define USART_RXC_vect USART_RX_vect
        #endif
        #ifdef USART0_RX_vect
            #define USART_RXC_vect USART1_RX_vect
        #endif        
    #else // Configured to use the first serial port
        #define UCSRA UCSR0A
        #define UCSRB UCSR0B
        #define UCSRC UCSR0C
        #define UBRRH UBRR0H
        #define UBRRL UBRR0L
        #define UDRE  UDRE0
        #define UDR   UDR0
        #define RXC   RXC0
        #define RXEN  RXEN0
        #define TXEN  TXEN0
        #define UCSZ0 UCSZ00
        #define UCSZ1 UCSZ01
        #define U2X   U2X0
        #define RXCIE RXCIE0
        #ifdef USART_RX_vect
            #define USART_RXC_vect USART_RX_vect
        #endif
        #ifdef USART0_RX_vect
            #define USART_RXC_vect USART0_RX_vect
        #endif
    #endif
#endif

#ifdef SERIALPORT   // Xmega controller

    // Write a character to the serial port
    static int serial_write(char c, FILE *f) {
        #if TERMINAL_MODE
            if (c=='\n') {
                // wait until transmitter is ready
                loop_until_bit_is_set(SERIALPORT.STATUS,USART_DREIF_bp); 
                SERIALPORT.DATA='\r';
            }
        #endif
        // wait until transmitter is ready
        loop_until_bit_is_set(SERIALPORT.STATUS,USART_DREIF_bp); 
        SERIALPORT.DATA = c;  
        return 0;
    }
    
    // Read a character from input buffer
    static int serial_read(FILE *f) {
        // wait until something is received
        while (inputStart==inputEnd);
        char c=inputBuffer[inputStart];
        if (++inputStart>=SERIAL_INPUT_BUFFER_SIZE) {
            inputStart=0;
        }
        return c;
    }
    
    // Initialize the serial port
    void initserial(void) {
        int bsel=(int)(4*(((float)F_CPU/(16*(float)SERIAL_BITRATE))-1));
        SERIALPORT.BAUDCTRLA = (uint8_t) bsel;
        SERIALPORT.BAUDCTRLB = (-2<<4) | (bsel >> 8); // BSCALE = -2
        SERIALPORT.CTRLC = USART_CHSIZE_8BIT_gc | USART_PMODE_DISABLED_gc;    // async, no parity, 8 bit data, 1 stop bit     
        SERIALPORT.CTRLA = USART_RXCINTLVL_HI_gc; // receive interrupts enable 
        PMIC.CTRL |= PMIC_HILVLEN_bm; // Enable high level interrupts
        SERIALPORT.CTRLB = USART_TXEN_bm | USART_RXEN_bm; // Enable transmitter and receiver  	
        sei(); 
        // Bind stdout and stdin to the serial port
        stdout = &serialPort;
        stdin = &serialPort;
        inputStart=0;
        inputEnd=0;
    }
    
    // Receive interrupt
    ISR ( SERIALPORT_RXC_vect ) {
        char c=SERIALPORT.DATA;
        #if TERMINAL_MODE
            // If received a backspace character, then discard the last received character
            if (c=='\b') {
                if (inputStart!=inputEnd) {
                    if (--inputEnd<0) {
                        inputEnd=0;
                    }                   
                }
                #if SERIAL_ECHO
                    serial_write(c,0);
                #endif 
                return;
            }
            // If received a line feed and the previous character was a 
            // carriage-return (or vice versa) then ignore the new character
            if ((c=='\n' && prevChar=='\r') || (c=='\r' && prevChar=='\n')) {
                return;
            }
            prevChar=c;
            // If received a carriage-return, then replace it by a line-feed 
            if (c=='\r') {
                c='\n';
            }
        #endif
        // Buffer the received character        
        inputBuffer[inputEnd]=c;
        if (++inputEnd>=SERIAL_INPUT_BUFFER_SIZE) {
            inputEnd=0;
        }        
        // Discard oldest buffered character in case of buffer overflow
        if (inputEnd==inputStart) {
            if (++inputStart>=SERIAL_INPUT_BUFFER_SIZE) {
                inputStart=0;
            }
        }  
        #if SERIAL_ECHO
            serial_write(c,0);
        #endif 
    }

#endif  // Xmega controller
       
#ifdef TXEN // ATmega controller

    #define BAUD SERIAL_BITRATE
    #include <util/setbaud.h>

    // Write a character to the serial port
    static int serial_write(char c, FILE *f) {
        #if TERMINAL_MODE
            if (c=='\n') {
                // wait until transmitter is ready
                loop_until_bit_is_set(UCSRA, UDRE);
                UDR='\r';
            }
        #endif
        // wait until transmitter is ready
        loop_until_bit_is_set(UCSRA, UDRE);
        UDR = c;
        return 0;
    }

    // Read a character from serial port
    static int serial_read(FILE *f) {
        // wait until something is received
        while (inputStart==inputEnd);         
        char c=inputBuffer[inputStart];
        if (++inputStart>=SERIAL_INPUT_BUFFER_SIZE) {
            inputStart=0;
        }
        return c;        
    }

    // Initialize the serial port
    void initserial(void) {
        // set baudrate
        UBRRH = UBRRH_VALUE;
        UBRRL = UBRRL_VALUE;
        #if USE_2X
            UCSRA |= (1 << U2X);
        #else
            UCSRA &= ~(1 << U2X);
        #endif
        // framing format 8N1
        #ifdef URSEL
            UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0);
        #else 
            UCSRC = (1<<UCSZ1) | (1<<UCSZ0);
        #endif
        // enable receiver and transmitter
        UCSRB = (1<<RXEN) | (1<<TXEN) | (1 << RXCIE);
        sei();
        // Bind stdout stdin to the serial port
        stdout = &serialPort;
        stdin = &serialPort;
        inputStart=0;
        inputEnd=0;
    }
    
    // USART Receive interrupt
    ISR ( USART_RXC_vect ) { 
        char c=UDR;
        #if TERMINAL_MODE
            // If received a backspace character, then discard the last 
            // received character
            if (c=='\b') {
                if (inputStart!=inputEnd) {
                    if (--inputEnd<0) {
                        inputEnd=0;
                    }               
                }
                #if SERIAL_ECHO
                    serial_write(c,0);
                #endif 
                return;
            }
            // If received a line feed and the previous character was a 
            // carriage-return (or vice versa) then ignore the new character
            if ((c=='\n' && prevChar=='\r') || (c=='\r' && prevChar=='\n')) {
                return;
            }
            prevChar=c;
            // If received a carriage-return, then replace it by a line-feed
            if (c=='\r') {
                c='\n';
            }
        #endif
        // Buffer the received character        
        inputBuffer[inputEnd]=c;
        if (++inputEnd>=SERIAL_INPUT_BUFFER_SIZE) {
            inputEnd=0;
        }        
        // Discard oldest buffered character in case of buffer overflow
        if (inputEnd==inputStart) {
            if (++inputStart>=SERIAL_INPUT_BUFFER_SIZE) {
                inputStart=0;
            }
        }  
        #if SERIAL_ECHO
            serial_write(c,0);
        #endif 
    }

#endif // ATmega controller
