/*
Minimalistic web server, that loads files from program memory (flash).
It supports include files and it can execute functions to generate
dynamic content.

The filesystem is in httpd-fsdata.c. All filenames
begin with a slash to indicate that they reside in the root
directory. Space characters are not allowed in filenames.
Subdirectories are not supported.

Template files for scripts have filename pattern *.shtml.
These files must be written with an 8-bit encoding, e.g.
iso8859-1. Within shtml files, two types of scripting markers
are supported:

%!: file_name              includes a html fragment file here.
%! function_name?params    executes a function

Functions can also be called via URL, such as
http://hostname/function?parameters. To call a function via URL that does not
use parameters, enter any dummy parameter, otherwise httpd thinks that
this is a regular filename and no function call. The function "io?params"
calls functions from user-functions.c or io-commands.c.

If the function name is the same as a filename, then the webbrowser
first executes that function, then returns the functions result string
and then returns the content of that file.

Include files cannot include other files. Nesting of include files is not
supported.

The size of GET request (from GET until end of parameters) must be smaller
than the inputbuffer. The inputbuffer size is defined in httpd-h.

*/

#include "../uip/uip.h"
#include "httpd.h"
#include "httpd-fs.h"
#include "httpd-functions.h"
#include "../inetd/inetd.h"
#include "../driver/CP2200.h"
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

#ifdef AUTH
#include "base64.h"
#endif // AUTH

#ifdef CARDSLOT
    #include "../avrfat32/mmc_config.h"
    #include "../avrfat32/file.h"
    #include "../avrfat32/fat.h"
    // these variables are located in the main.c file.
    extern uint8_t card_present; // set to 1 during start-up if the SD card is accessible
    extern uint8_t card_busy;    // Used to ensure that only one single thread uses the avrfat32 library
#endif

const char method_get[]        PROGMEM="GET ";
const char nameof_index_file[] PROGMEM="/index.shtml";
const char nameof_404_html[]   PROGMEM="/404.html";

// Filename suffixes, used by send_headers to determine content type

const char suffix_html[]  PROGMEM=".html";
const char suffix_shtml[] PROGMEM=".shtml";
const char suffix_css[]   PROGMEM=".css";
const char suffix_png[]   PROGMEM=".png";
const char suffix_gif[]   PROGMEM=".gif";
const char suffix_jpg[]   PROGMEM=".jpg";
const char suffix_jpeg[]  PROGMEM=".jpeg";
const char suffix_txt[]   PROGMEM=".txt";
const char suffix_xml[]   PROGMEM=".xml";

// HTTP content types, used by send_headers

const char type_html[]    PROGMEM="Content-Type: text/html\r\n\r\n"; // HTML files control chache via embedded headers
const char type_css[]     PROGMEM="Content-Type: text/css\r\nCache-Control: max-age=3600\r\n\r\n";
const char type_png[]     PROGMEM="Content-Type: image/png\r\nCache-Control: max-age=3600\r\n\r\n";
const char type_gif[]     PROGMEM="Content-Type: image/gif\r\nCache-Control: max-age=3600\r\n\r\n";
const char type_jpeg[]    PROGMEM="Content-Type: image/jpeg\r\nCache-Control: max-age=3600\r\n\r\n";
const char type_xml[]     PROGMEM="Content-Type: text/xml\r\nCache-Control: no-cache\r\n\r\n";
const char type_plain[]   PROGMEM="Content-Type: text/plain\r\nCache-Control: no-cache\r\n\r\n";
const char type_unknown[] PROGMEM="Cache-Control: no-cache\r\n\r\n";

// HTTP Status headers
const char status_200[]   PROGMEM="HTTP/1.1 200 OK\r\nConnection: close\r\n";
const char status_404[]   PROGMEM="HTTP/1.1 404 Not found\r\nConnection: close\r\n";

#ifdef AUTH
// HTTP Header for authorisation
const char auth_header[]  PROGMEM="Authorization: Basic "; // followed by the base64 encrypted authentication key, followed by a line break
const char status_401[]   PROGMEM="HTTP/1.1 401 Auth Required\r\nWWW-Authenticate: Basic realm=\"Controller\"\r\nConnection: close\r\n\r\nAuth Required\r\n";
#endif // AUTH


//------------------------------------------------------------------------------------------------

// Send HTTP headers, depending on the filename suffix
// If content_length is 0, no such header will be sent
static PT_THREAD(send_headers(struct application_state *app_state, const char *statushdr, const uint32_t content_length)) {
    PSOCK_BEGIN(&app_state->socket);
    // Send the HTTP status header
    PSOCK_SEND_STR_P(&app_state->socket, statushdr);

    #ifdef CARDSLOT
        if (content_length>0) {
            // Send a content-length header
            PSOCK_SEND_STR_P(&app_state->socket, PSTR("Content-Length: "));
            sprintf(app_state->mmc_filesize,"%lu\r\n",file.length);
            PSOCK_SEND_STR(&app_state->socket,app_state->mmc_filesize);
        }
    #endif

    char* ptr = strrchr(app_state->filename, '.');
    if (ptr == NULL) {
        PSOCK_SEND_STR_P(&app_state->socket, type_unknown);
    } else if (strcmp_P(ptr, suffix_html) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_html);
    } else if (strcmp_P(ptr, suffix_shtml) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_html);
    } else if (strcmp_P(ptr, suffix_css) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_css);
    } else if (strcmp_P(ptr, suffix_png) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_png);
    } else if (strcmp_P(ptr, suffix_gif) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_gif);
    } else if (strcmp_P(ptr, suffix_jpg) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_jpeg);
    } else if (strcmp_P(ptr, suffix_jpeg) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_jpeg);
    } else if (strcmp_P(ptr, suffix_txt) == 0) {
        PSOCK_SEND_STR_P(&app_state->socket, type_plain);
    } else {
        PSOCK_SEND_STR_P(&app_state->socket, type_unknown);
    }
    PSOCK_END(&app_state->socket);
}


//---------------------------------------------------------------------------------------

#ifdef CARDSLOT
    // Send the content of an SD/MMC file that has been opened before calling this function

    static PT_THREAD(send_mmc_file(struct application_state *app_state)) {
        PSOCK_BEGIN(&app_state->socket);
        app_state->mmc_file_remaining=file.length;
        do {
            // Read one sector from the SD/MMC card.
            // I also tried to send packets of 2*512 bytes but that leaded to worse performance.
            ffread();
            file.cntOfBytes=512;

            // How many bytes have been read from the file?
            unsigned short bytes=app_state->mmc_file_remaining>512?512:app_state->mmc_file_remaining;

            // Send the bytes
            PSOCK_SEND(&app_state->socket,(char*)fat.sector,bytes);

            app_state->mmc_file_remaining-=bytes;
        }
        while (app_state->mmc_file_remaining > 0);
        PSOCK_END(&app_state->socket);
    }
#endif // CARDSLOT

//--------------------------------------------------------------------------
#if defined CARDSLOT && (MMC_LS==TRUE)

    // Special version of lsRowsOfClust2 from vfat32, that writes directory into the ethernet buffer.
    // List the filenames of one cluster. Each cluster contains up to 16 filenames.

    static int lsRowsOfClust2 (uint32_t start_sec) {
        uint16_t row; // rows
        uint8_t sec=0; // sectors
        uint8_t tmp[13]; // temp. buffer for filename
        int written=0; // how many characters have been written into the buffer
        do {
            fat_loadSector( start_sec + sec ); // load sectors of the cluster
            for( row=0; row<512; row+=32 ) {  // go through the rows of the sector
                if( (fat.sector[row+11]==0x20
                    /*dont display dirs ||fat.sector[row+11]==0x10*/
                  ) && (fat.sector[row]!=0xE5 && fat.sector[row]!=0x00) ) {
                    // extract and format filename
                    int j=0;
                    for( uint8_t i=0 ; i<11 ; i++ ){
                        uint8_t c=fat.sector[row+i];
                        if (c!=0 && c!=' ') tmp[j++]=c;
                        if (i==7) tmp[j++]='.';
                    }
                    tmp[j]='\0';
                    // load row into the file struct to get file.length
                    fat_loadRowOfSector(row);
                    // output filename and size into the buffer
                    if (uip_mss()>=56*16) {
                      written+=sprintf(uip_appdata+written,"<a href=\"SD/%s\">%12s</a> (%10li)\n",tmp,tmp,file.length);
                    }
                    else {
                      // shorter version for small ethernet buffer
                      written+=sprintf(uip_appdata+written,"%12s %10li\n",tmp,file.length);
                    }
                   
                }
            }
        }
        while( ++sec < fat.secPerClust );
        return written;
    }

    // Generate a listing of the SD/MMC card content.
    // This function is called repeatedly to output only one cluster of 16 filenames
    // in each repetition. 16 filenames always fit into the ethernet packet buffer.
    // Returns the number of bytes that have been copied into the ethernet buffer.
    static unsigned short generate_mmc_directory(void *app_state) {
        struct application_state *state = (struct application_state *) app_state;
        uint8_t  sectors; // count sectors
        uint32_t clusters; // clusters of the current directory
        int current_cluster=0;  // number of the current cluster. Counts up while processing the TOC.

        // find out which is the startting cluster, depends on if the current dir is the root dir.
        clusters = (fat.dir==0) ? fat.rootDir:fat.dir;

        // output root dir of fat16
        if(fat.dir==0 && fat.fatType==16){
            sectors = 0;
            do {
                current_cluster++;
                // output only the requested cluster
                if (current_cluster==state->mmc_next_cluster) {
                    state->mmc_next_cluster++;
                    return lsRowsOfClust2(clusters + sectors );
                }
                sectors += fat.secPerClust;
            }
            while( sectors < (uint8_t)32 );
        }

        // output root dir of fat32 or any other dir of fat16/32
        else {
            // walk through the clusters of the dir
            while(!((clusters>=0x0ffffff8&&fat.fatType==32)||(clusters>=0xfff8&&fat.fatType==16))) { // are there more sectors?
                current_cluster++;
                // output only the requested cluster
                if (current_cluster==state->mmc_next_cluster) {
                    return lsRowsOfClust2(fat_clustToSec(clusters) );
                }
                clusters = fat_getNextCluster(clusters);
            }
        }

        // end of directory reached.
        return 0;
    }


    // List the content of the SD/MMC card.
    // Wait until the card is ready, if necessary.
    static PT_THREAD(list_mmc_directory(struct application_state *app_state)) {
        PSOCK_BEGIN(&app_state->socket);

        if (card_present==1) {

            // Synchronize access to the SD/MMC card
            if (card_busy==1) {
                PSOCK_WAIT_UNTIL(&app_state->socket, card_busy==0);
            }
            card_busy=1;
            app_state->mmc_locked=1;

            // start with the first cluster
            app_state->mmc_next_cluster=1;
            do {
                PSOCK_GENERATOR_SEND(&app_state->socket, generate_mmc_directory, app_state);
                app_state->mmc_next_cluster++;
            }
            while (app_state->socket.sendlen >0 );
        }
        else {
            PSOCK_SEND_STR_P(&app_state->socket, PSTR("no card"));
        }
        PSOCK_END(&app_state->socket);
    }

#endif // CARDSLOT

// --------------------------------------------------------------------------------
// Functions for sending files from Program memory

// copy the max. allowed number of bytes from file to the output
// buffer uip_appdata.
// After sending, app_state->partlen contains the number of bytes that have been sent,
// which might be less than requested.

static unsigned short generate_part_of_file_P(void *app_state) {
    struct application_state *state = (struct application_state *) app_state;
    if (state->file.len > uip_mss()) {
        state->partlen = uip_mss();
    } else {
        state->partlen = state->file.len;
    }
    memcpy_P(uip_appdata, state->file.data, state->partlen);
    return state->partlen;
}


// Send the whole file app_state->file

static PT_THREAD(send_file_P(struct application_state *app_state)) {
    PSOCK_BEGIN(&app_state->socket);
    do {
        PSOCK_GENERATOR_SEND(&app_state->socket, generate_part_of_file_P, app_state);
        // Move the file pointer forward to prepare for sending the next part
        app_state->file.len -= app_state->partlen;
        app_state->file.data += app_state->partlen;
    }
    while (app_state->file.len > 0);
    PSOCK_END(&app_state->socket);
}


// Send app_state->partlen bytes of a file starting at app_state->file.data.
// After sending, app_state->partlen contains the number of bytes that have been sent,
// which might be less than requested.

static PT_THREAD(send_part_of_file_P(struct application_state *app_state)) {
    PSOCK_BEGIN(&app_state->socket);
    PSOCK_SEND_P(&app_state->socket, app_state->file.data, app_state->partlen);
    PSOCK_END(&app_state->socket);
}

//--------------------------------------------------------------------------

// decode an URL parameter. The result is stored where the source came from.
// The parameter must be terminated with an ascii 0 or an & character.

void url_decode(char* parameter) {
    char high_nibble;
    char low_nibble;
    while (parameter[0]!=0) {
         if (parameter[0]=='+')
             parameter[0]=' ';
         else if (parameter[0]=='%') {
             // convert the hexadecimal number that follows the "%" to char
             if (parameter[1]<='9')
                 high_nibble=parameter[1]-'0';
             else
                 high_nibble=parameter[1]-'A'+10;
             if (parameter[2]<='9')
                 low_nibble=parameter[2]-'0';
             else
                 low_nibble=parameter[2]-'A'+10;
             char c=(high_nibble<<4)+low_nibble;
             // replace the "%" char by the new character
             parameter[0]=c;
             //shift the remaining string 2 characters forward.
             memmove(parameter+1,parameter+3,strlen(parameter+3)+1);
         }
         else if (parameter[0]=='&') {
            parameter[0]=0;
            break;
         }
         parameter++;
    }
}



// Call an i/o function and send its result
// Note that the function may be called repeatedly in case of TCP re-transmissions

static PT_THREAD(call_function(struct application_state *app_state)) {
    PSOCK_BEGIN(&app_state->socket);
    // call the function dispatcher from httpd-functions.c
    PSOCK_GENERATOR_SEND(&app_state->socket, httpd_function_dispatcher, app_state);
    PSOCK_END(&app_state->socket);
}



// Find the end of the current script marker (including function name and parameters).
// len is the maximum number of bytes to search in.
// Returns the index of the next line break or the string end.

static int end_of_script_marker_P(char* data, int len) {
    int i = 0;
    do {
        char c=pgm_read_byte(data + i);
        if (c=='\n' || c=='\r' || c==' ' || c=='"' || c=='<' || c==0)
            return i;
        i++;
    }
    while (i < len);
    return len;
}


// Find the next script marker in file data.
// len is the maximum number of bytes to search in.
// Returns the index of the nextscript marker or -1.

static int next_scriptmarker_P(char* data, int len) {
    int i = 0;
    len--;
    do {
        if (pgm_read_byte(data + i) == '%' && pgm_read_byte(data + i + 1) == '!')
            return i;
        i++;
    }
    while (i < len);
    return -1;
}


// process the shtml file app_state->file
// This is done in a loop that splits the file data into parts, where the delimiters
// are script markers.

static PT_THREAD(handle_script(struct application_state *app_state)) {
    PT_BEGIN(&app_state->scriptpt);
    // repeat until end of file data
    while (app_state->file.len > 0) {
        // check if the data start with a script marker
        int marker1 = next_scriptmarker_P(app_state->file.data, 1);
        if (marker1 >= 0) {
            // check if this is an include command
            if (pgm_read_byte(app_state->file.data + 2) == ':') {
                // it is an include file
                // safe the current file data pointer+length because sending an include file will overwrite them.
                app_state->file_backup.data = app_state->file.data;
                app_state->file_backup.len = app_state->file.len;
                app_state->file_backup.index = app_state->file.index;
                // open and send the include file
                int index=httpd_fs_open_P(app_state->file.data + 4, &app_state->file); // filename,destination
                if (index>=0) {
                    PT_WAIT_THREAD(&app_state->scriptpt, send_file_P(app_state));
                    // count access to the file that has been sent
                    httpd_fs_countAccess(app_state->file.index);
                }
                // restore the saved data pointer+length
                app_state->file.data = app_state->file_backup.data;
                app_state->file.len = app_state->file_backup.len;
                app_state->file.index = app_state->file_backup.index;
            }
            else {
                // it is a function call
                // Copy function name and parameters to the input buffer and
                // set the filename and parameters pointers to the correct positions
                // within inputbuffer and terminate these strings with a 0.
                app_state->filename=app_state->inputbuffer;
                app_state->parameters=NULL;
                int i=0;
                while (1) {
                    char c=pgm_read_byte(app_state->file.data+3+i);
                    if (c=='?') {
                         app_state->parameters=app_state->inputbuffer+i+1;
                         app_state->inputbuffer[i]=0;
                    }
                    else if (c==' ' || c=='\r' || c=='\n' || c=='"' || c=='<' ||c==0) {
                        app_state->inputbuffer[i]=0;
                        break;
                    }
                    else {
                        app_state->inputbuffer[i]=c;
                    }
                    i++;
                }
                if (app_state->parameters==NULL) {
                    app_state->parameters=app_state->inputbuffer+i;
                }

                #if defined CARDSLOT && (MMC_LS==TRUE)
                    // if the filename is "/ls" then display the content of the SD/MMC card
                    if (strcmp(app_state->filename,"ls")==0) {
                        PT_WAIT_THREAD(&app_state->scriptpt,list_mmc_directory(app_state));
                    }
                    else
                #endif

                {
                    // call the function dispatcher
                    PT_WAIT_THREAD(&app_state->scriptpt, call_function(app_state));
                }
            }
            // restore the saved data pointer+length
            // and move the file pointer to the end of the current script tag
            int end = end_of_script_marker_P(app_state->file.data+4, app_state->file.len-4);
            app_state->file.data+=end+4;
            app_state->file.len-=end+4;

        } else {
            // the data does not start with a script marker

            // see if the data contains a script marker somewhere else
            int marker2 = next_scriptmarker_P(app_state->file.data, app_state->file.len);
            if (marker2 >= 0) {
                // limit the part length to the max. allowed size (uip_mss)
                if (marker2 > uip_mss()) {
                    app_state->partlen = uip_mss();
                    // We decrease uip_mss by 8 to improve performance. This hack is described in the uIP documentation.
                } else {
                    app_state->partlen = marker2;
                }
                // send the part that comes before the next script mark
                PT_WAIT_THREAD(&app_state->scriptpt, send_part_of_file_P(app_state));
                // move the file pointer forward to prepare sending the next part.
                app_state->file.data += app_state->partlen;
                app_state->file.len -= app_state->partlen;
            } else {
                // there is no script marker anymore
                // send the whole file
                PT_WAIT_THREAD(&app_state->scriptpt, send_file_P(app_state));
                app_state->file.len = 0; // reached end of file, end the loop
            }
        }
    }
    PT_END(&app_state->scriptpt);
}


// -----------------------------------------------------------------------------
// Read the incoming HTTP request "GET /filename?parameters HTTP/1.1..."

static PT_THREAD(get_request(struct application_state *app_state)) {
    PSOCK_BEGIN(&app_state->socket);

    // read until the first space to get the HTTP request method
    PSOCK_READTO(&app_state->socket, ' ');
    //if this is not a GET request, close connection and exit
    if (strncmp_P(app_state->inputbuffer, method_get, sizeof(method_get)-1) != 0) {
        PSOCK_CLOSE_EXIT(&app_state->socket);
    }
    // read until the second space to get the filename and parameters
    PSOCK_READTO(&app_state->socket, ' ');

    // if the filename is "/", then replace it by the name of the index file
    if (app_state->inputbuffer[0] == '/' && app_state->inputbuffer[1] == ' ') {
        strcpy_P(app_state->inputbuffer, nameof_index_file);
    }

    // Set the filename and parameters pointers to the correct positions
    // withing inputbuffer and terminate these strings with a 0.
    app_state->filename=app_state->inputbuffer;
    app_state->parameters=NULL;
    int i=0;
    while (1) {
        char c=app_state->inputbuffer[i];
        if (c=='?') {
            app_state->parameters=app_state->inputbuffer+i+1;
            app_state->inputbuffer[i]=0;
        }
        else if (c==' ' || c==0) {
            app_state->inputbuffer[i]=0;
            break;
        }
        i++;
    }
    if (app_state->parameters==NULL) {
        app_state->parameters=app_state->inputbuffer+i;
    }

    #ifdef AUTH
        if (config.authentication[0]==0) {
            app_state->isAuthenticated=1;
        }
        else {
            app_state->isAuthenticated=0;
            // back-up the filename and params
            memcpy(app_state->authbuffer,app_state->inputbuffer,sizeof(app_state->authbuffer));
            // read the HTTP headers until an empty line, which marks the end of the headers
            do {
                PSOCK_READTO(&app_state->socket, '\n');
                // check if we got an authorisation header
                if (strncmp_P(app_state->inputbuffer, auth_header, sizeof(auth_header)-1) == 0) {
                    char* receivedpw=app_state->inputbuffer+sizeof(auth_header)-1;
                    // remove line break and space after received key
                    int i=0;
                    while (receivedpw[i]!='\n' && receivedpw[i]!='\r' && receivedpw[i]!=' ') {
                    i++;
                    }
                    receivedpw[i]=0;
                    // We store the encoded authentication key at the end of the input buffer,
                    // where we have some free space. This reduces RAM usage.
                    char* encodedpw=app_state->inputbuffer+60;
                    encode64(encodedpw,config.authentication);
                    // compare the received and the encoded configured keys.
                    if (strcmp(encodedpw,receivedpw)==0) {
                        app_state->isAuthenticated=1;
                    }
                    break; // skip all following headers, we are not interested in them
                }
            }
            while (app_state->inputbuffer[0]!='\r' && app_state->inputbuffer[0]!='\n'); // empty line
            // recover the inputbuffer which holds filename and parameters
            memcpy(app_state->inputbuffer,app_state->authbuffer,sizeof(app_state->inputbuffer));
        }
    #endif // AUTH

    PSOCK_END(&app_state->socket);
}

//------------------------------------------------------------------------------------
// Receive a HTTP request and generate a response

static PT_THREAD(not_authenticated(struct application_state *app_state)) {
    PSOCK_BEGIN(&app_state->socket);
    // send a HTTP 401 header = Auth Required
    PSOCK_SEND_STR_P(&app_state->socket, status_401);
    PSOCK_END(&app_state->socket);
}

static PT_THREAD(handle_connection(struct application_state *app_state)) {
    PT_BEGIN(&app_state->outputpt);

    // First receive the incoming HTTP request
    PT_WAIT_THREAD(&app_state->outputpt, get_request(app_state));

#ifdef AUTH
    // check if the caller is authenticated
    if (app_state->isAuthenticated==0) {
        // send a HTTP 401 header = Auth Required
        PT_WAIT_THREAD(&app_state->outputpt, not_authenticated(app_state));
    }
    else
#endif // AUTH
    {

        #ifdef CARDSLOT
            // if the filename start with "/SD/" then load a file from the SD/MMC card
            if (strncmp_P(app_state->filename,PSTR("/SD/"),4)==0) {
                app_state->filename+=4;

                uint8_t file_status=MMC_FILE_ERROR;
                if (card_present==1) {

                    // Synchronize access t the SD/MMC card
                    if (card_busy==1) {
                        PT_WAIT_UNTIL(&app_state->outputpt, card_busy==0);
                    }
                    card_busy=1;
                    app_state->mmc_locked=1;

                    file_status=ffopen((uint8_t*)app_state->filename,'r');
                }
                if (file_status==MMC_FILE_OPENED) {
                    // send a HTTP header
                    PT_WAIT_THREAD(&app_state->outputpt, send_headers(app_state, status_200, file.length));
                    // send the file content
                    PT_WAIT_THREAD(&app_state->outputpt, send_mmc_file(app_state));
                }
                else {
                    // file not found
                    httpd_fs_open_P(nameof_404_html, &app_state->file);
                    strcpy_P(app_state->filename, nameof_404_html);
                    PT_WAIT_THREAD(&app_state->outputpt, send_headers(app_state, status_404,0));
                    PT_WAIT_THREAD(&app_state->outputpt, send_file_P(app_state));
                }

            }

            #if MMC_LS==TRUE
                // if the filename is "/ls" then display the content of the SD/MMC card
                else if (strcmp(app_state->filename,"/ls")==0) {
                    PT_WAIT_THREAD(&app_state->outputpt, send_headers(app_state, status_200,0));
                    // List content of the current directory
                    PT_WAIT_THREAD(&app_state->outputpt, list_mmc_directory(app_state));
                }
            #endif

            else
        #endif // CARDSLOT

        {
            // try to load the file
            httpd_fs_open(app_state->filename, &app_state->file);

            // If we have parameters, then this is a direct function call
            if (app_state->parameters[0]!=0) {
                // send a HTTP header
                PT_WAIT_THREAD(&app_state->outputpt, send_headers(app_state, status_200,0));
                // The function name is the filename but without the first slash.
                app_state->filename++;
                // Call the function dispatcher
                PT_WAIT_THREAD(&app_state->outputpt, call_function(app_state));

                // restore the original filename
                app_state->filename--;

                // if there is a file with the same name as the function, then load and display it.
                if (app_state->file.index >=0 ) {
                    // If the file is an .shtml file, then spawn a new sub thread that
                    // processes the script file.
                    char* ptr = strchr(app_state->filename, '.');
                    if (ptr != NULL && strcmp_P(ptr, suffix_shtml) == 0) {
                        PT_INIT(&app_state->scriptpt);
                        PT_WAIT_THREAD(&app_state->outputpt, handle_script(app_state));
                    }
                    else {
                        // It's not a script file, send the file
                        PT_WAIT_THREAD(&app_state->outputpt, send_file_P(app_state));
                    }
                    // count access to the file that has been sent
                    httpd_fs_countAccess(app_state->file.index);
                }
            }
            else {
                // If we dont have parameters, it is not a function call.
                // If the file does not exists, return HTTP error page
                if (app_state->file.index < 0) {
                    httpd_fs_open_P(nameof_404_html, &app_state->file);
                    strcpy_P(app_state->filename, nameof_404_html);
                    PT_WAIT_THREAD(&app_state->outputpt, send_headers(app_state, status_404,0));
                    PT_WAIT_THREAD(&app_state->outputpt, send_file_P(app_state));
                }
                else {
                    // File exists, send HTTP status 200 and filename specific header
                    PT_WAIT_THREAD(&app_state->outputpt, send_headers(app_state, status_200,0));
                    // If the file is an .shtml file, then spawn a new sub thread that
                    // processes the script file.
                    char* ptr = strchr(app_state->filename, '.');
                    if (ptr != NULL && strcmp_P(ptr, suffix_shtml) == 0) {
                        PT_INIT(&app_state->scriptpt);
                        PT_WAIT_THREAD(&app_state->outputpt, handle_script(app_state));
                    }
                    else {
                        // It's not a script file, send the file
                        PT_WAIT_THREAD(&app_state->outputpt, send_file_P(app_state));
                    }
                }
                // count access to the file that has been sent
                httpd_fs_countAccess(app_state->file.index);
            }
        }
    }

    // close the connection
    uip_close();
    PT_END(&app_state->outputpt);
}


// Process application events. This is called by uIP whenever an event related to
// the connection occurs and also periodically

void httpd_appcall(struct application_state *app_state) {
    // is the connection closed?
    if (uip_closed() || uip_aborted() || uip_timedout()) {
        #ifdef CARDSLOT
            // release the lock of the SD/MMC card when the connection has been closed or aborted
            if(app_state->mmc_locked==1) {
                ffclose(); // no matter if there is no open file at the moment
                card_busy=0;
                app_state->mmc_locked=0;
            }
        #endif
        return;
    }
    // is this a new connection?
    if (uip_connected()) {
        // Initialize a Protosocket and a Protothread
        PSOCK_INIT(&app_state->socket, app_state->inputbuffer, sizeof (app_state->inputbuffer));
        PT_INIT(&app_state->outputpt);
        #ifdef CARDSLOT
            app_state->mmc_locked=0;  // will be set to 1 when the webserver locks the mmc card for exclusive access
        #endif
    }
    // process events for an open connection
    handle_connection(app_state);
}



