#ifndef _HW_LAYOUT_H__
#define _HW_LAYOUT_H__


// This file is used to adapt the firmware to different hardware layouts

// ADBUS     parallel A/D bus
// ALE       address latch enable of parallel A/D bus
// nWr       write data on parallel A/D bus
// nRd       read data on parallel A/D bus

// nCSETH    chip select of ethernet controller
// nCSADC    chip select of ADC chip
// nCSMMC    chip select of the SD/MMC card slot (optional)

// nRSTETH   reset of ethernet controller (optional, must be driven in open-drain mode)

// DIN       data input of serial ADC chip
// DOUT      data output of serial ADC chip
// CLK       clock line of serial ADC chip


// Macro used to write to a single I/O pin
#define writeBit(port,bit,value) { if ((value)>0) (port) |= (1<<bit); else (port) &= ~(1<<bit); } 

// Macro used to read from a single I/O pin
#define readBit(port,bit) (((port) >> (bit)) & 1)

// Crumb644-NET version 1.0 from www.chip45.com

#ifdef Crumb644_NET_v1
    // Macros for access to address/data bus
    #define ADBUS_output()       DDRA = 255
    #define ADBUS_input()        DDRA = 0 
    #define ADBUS_write(value)   PORTA = (value)
    #define ADBUS_read()         PINA
    
    // Macros for the control lines of the adress/data bus
    #define ALE(value)           writeBit(PORTC,3,value)
    #define nWR(value)           writeBit(PORTC,4,value)
    #define nRD(value)           writeBit(PORTC,5,value)
    
    // Macros for chip-select signals
    #define nCSADC(value)        writeBit(PORTC,2,value)
    #define nCSETH(value)        writeBit(PORTC,6,value)
       
    // The status LED is accessed in open-drain mode
    #define STATUS_LED_on()      { \
                                   writeBit(PORTC,0,0); \
                                   writeBit(DDRC,0,1); \
                                 }
    #define STATUS_LED_off()     writeBit(DDRC,0,0)
    
    // Initialisation of the IO pins for ethernet, status LED's and Reset jumper.
    #define init_io_pins()       { \
                                   DDRA=0; \
                                   DDRC=(1<<2)+(1<<3)+(1<<4)+(1<<5)+(1<<6); \
                                   PORTC=(1<<2)+(1<<3)+(1<<4)+(1<<5)+(1<<6); \
                                 }   
                                 
    // Config reset jumper on ISP connector.
    // PB5 is used as input with pull-ups and PB7 is used as output
    #define RST_JUMPER_init()           { \
                                           writeBit(DDRB,7,1); \
                                           writeBit(DDRB,5,0); \
                                           writeBit(PORTB,5,1); \
                                        }
    #define RST_JUMPER_write(value)     writeBit(PORTB,7,value)
    #define RST_JUMPER_read             readBit(PINB,5)
    #define RST_JUMPER_exit()           { \
                                          writeBit(DDRB,7,0); \
                                          writeBit(PORTB,5,0); \
                                        }
    
    // ADC hardware settings.
    #define ADC_CHANNELS         4
    
    // Serial ADC communication to MCP3204 chip.
    // PB5 is the data line from AVR to ADC
    // PB6 is the data line from ADC to AVR
    // PB7 is the clock line
    #define ADC_PORT             PORTB
    #define ADC_DDR              DDRB
    #define ADC_CLK(value)       writeBit(PORTB,7,value)
    #define ADC_write(value)     writeBit(PORTB,5,value)
    #define ADC_read()           readBit(PINB,6)
    #define init_adc_pins()      { \
                                   writeBit(DDRB,5,1); \
                                   writeBit(DDRB,6,0); \
                                   writeBit(PORTB,7,0); \
                                   writeBit(DDRB,7,1); \
                                 }
                                 
    // Settings for the Email application
    #ifdef EMAIL_APP
       #define email_PORT       PIND
    #endif  // EMAIL_APP
                                 
    // Definition of i/o port bits that are write-protected.
    // These pins cannot be accessed via io-commands d, p and o.
    // 0x00 means that the whole port is not protected.
    #define protect_PORTA        0xFF
    #define protect_PORTB        0x00
    #define protect_PORTC        0xFC
    #define protect_PORTD        0x00
#endif


// Crumb644-NET version 2.0 or 2.1 from www.chip45.com

#ifdef Crumb644_NET_v2
    // Macros for access to address/data bus
    #define ADBUS_output()       DDRA = 255
    #define ADBUS_input()        DDRA = 0 
    #define ADBUS_write(value)   PORTA = (value)
    #define ADBUS_read()         PINA
    
    // Control signals to the ethernet chip
    #define ALE(value)           writeBit(PORTC,3,value)
    #define nWR(value)           writeBit(PORTC,4,value)
    #define nRD(value)           writeBit(PORTC,5,value)
    #define nCSETH(value)        writeBit(PORTC,6,value)
       
    // The status LED is accessed in open-drain mode
    #define STATUS_LED_on()      { \
                                   writeBit(PORTC,0,0); \
                                   writeBit(DDRC,0,1); \
                                 }
    #define STATUS_LED_off()     writeBit(DDRC,0,0)
    
    // Initialisation of the IO pins for ethernet, status LED's and Reset jumper.
    #define init_io_pins()       { \
                                   DDRA=0; \
                                   DDRC=(1<<3)+(1<<4)+(1<<5)+(1<<6); \
                                   PORTC=(1<<3)+(1<<4)+(1<<5)+(1<<6); \
                                 }   
 
    // Config reset jumper on ISP connector.
    // PB5 is used as input with pull-ups and PB7 is used as output
    #define RST_JUMPER_init()           { \
                                           writeBit(DDRB,7,1); \
                                           writeBit(DDRB,5,0); \
                                           writeBit(PORTB,5,1); \
                                        }
    #define RST_JUMPER_write(value)     writeBit(PORTB,7,value)
    #define RST_JUMPER_read             readBit(PINB,5)
    #define RST_JUMPER_exit()           { \
                                          writeBit(DDRB,7,0); \
                                          writeBit(PORTB,5,0); \
                                        }

    // The ADC inputs are not accessible on this harware.
    #define ADC_CHANNELS         0
    
    // Settings for the Email application
    #ifdef EMAIL_APP
       #define email_PORT       PIND
    #endif  // EMAIL_APP

    // Definition of i/o port bits that are protected.
    // These pins cannot be accessed via io-commands d, p and o.
    // 0x00 means that the whole port is not protected.
    #define protect_PORTA        0xFF
    #define protect_PORTB        0x00
    #define protect_PORTC        0xFC
    #define protect_PORTD        0x00    
#endif


// Crumb644-NET version 2.0 or 2.1 from www.chip45.com
//
// with 1-4 shift registers 74HC165 for 8-32 additional inputs
// PD2 = Shift pulse (Pin 2 CP)
// PD3 = Data in     (Pin 9 Q7)
// PD4 = Load pulse  (Pin 1 nPL)
//
// with 1-4 shift registers 74HC595 for 8-32 additional outputs
// PD5 = Store pulse (Pin 12 STcp)
// PD6 = Shift pulse (Pin 11 SHcp)
// PD7 = Data out    (Pin 14 DS)

#ifdef Crumb644_NET_v2_XL
    // Macros for access to address/data bus
    #define ADBUS_output()       DDRA = 255
    #define ADBUS_input()        DDRA = 0 
    #define ADBUS_write(value)   PORTA = (value)
    #define ADBUS_read()         PINA
    
    // Control signals to the ethernet chip
    #define ALE(value)           writeBit(PORTC,3,value)
    #define nWR(value)           writeBit(PORTC,4,value)
    #define nRD(value)           writeBit(PORTC,5,value)
    #define nCSETH(value)        writeBit(PORTC,6,value)
       
    // The status LED is accessed in open-drain mode
    #define STATUS_LED_on()      { \
                                   writeBit(PORTC,0,0); \
                                   writeBit(DDRC,0,1); \
                                 }
    #define STATUS_LED_off()     writeBit(DDRC,0,0)   
       
    // Initialisation of the IO pins for ethernet, status LED's and Reset jumper.
    #define init_io_pins()       { \
                                   DDRA=0; \
                                   DDRC=(1<<3)+(1<<4)+(1<<5)+(1<<6); \
                                   PORTC=(1<<3)+(1<<4)+(1<<5)+(1<<6); \
                                   DDRD=(1<<2)+(1<<4)+(1<<5)+(1<<6)+(1<<7); \
                                   PORTD=(1<<4); \
                                 }   
                                 
    // Config reset jumper on ISP connector, PB=output, PB5=input with pull-up
    #define RST_JUMPER_init()           { \
                                          writeBit(DDRB,7,1); \
                                          writeBit(DDRB,5,0); \
                                          writeBit(PORTB,5,1); \
                                        }
    #define RST_JUMPER_write(value)     writeBit(PORTB,7,value)
    #define RST_JUMPER_read             readBit(PINB,5)
    #define RST_JUMPER_exit()           { \
                                          writeBit(DDRB,7,0); \
                                          writeBit(PORTB,5,0); \
                                        }

    // The ADC inputs are not accessible on this harware.
    #define ADC_CHANNELS         0
    
    // Settings for the Email application
    #ifdef EMAIL_APP
       #define email_PORT       PIND
    #endif  // EMAIL_APP

    // Definition of i/o port bits that are protected.
    // These pins cannot be accessed via io-commands d, p and o.
    // 0x00 means that the whole port is not protected.
    #define protect_PORTA        0xFF
    #define protect_PORTB        0x00
    #define protect_PORTC        0xFC
    #define protect_PORTD        0xFC
    
    // Number of shift register inputs and outputs
    #define XL_INPUTS            24
    #define XL_OUTPUTS           24
    
    // Macros for additional inputs through shift registers (Port X)
    // Some commands are duplicated to slow down the serial interface a little
    #define SHIFTin              { \
                                   writeBit(PORTD,2,1); \
                                   writeBit(PORTD,2,1); \
                                   writeBit(PORTD,2,0); \
                                   writeBit(PORTD,2,0); \
                                 }
    #define DATAin               readBit(PIND,3)
    #define LOADin               { \
                                   writeBit(PORTD,4,0); \
                                   writeBit(PORTD,4,0); \
                                   writeBit(PORTD,4,1); \
                                   writeBit(PORTD,4,1); \
                                 } 
    
    // Macros for additional outputs through shift registers (Port Y)
    // Some commands are duplicated to slow down the serial interface a little
    #define STOREout             { \
                                   writeBit(PORTD,5,1); \
                                   writeBit(PORTD,5,1); \
                                   writeBit(PORTD,5,0); \
                                   writeBit(PORTD,5,0); \
                                 }
    #define SHIFTout             { \
                                   writeBit(PORTD,6,1); \
                                   writeBit(PORTD,6,1); \
                                   writeBit(PORTD,6,0); \
                                   writeBit(PORTD,6,0); \
                                 }
    #define DATAout(value)       writeBit(PORTD,7,value)
#endif


// CrumbX1-NET version 1.0 or 1.1 from www.chip45.com

#ifdef CrumbX1_NET
    // Macros for access to address/data bus
    #define ADBUS_output()       PORTF.DIR = 255
    #define ADBUS_input()        PORTF.DIR = 0 
    #define ADBUS_write(value)   PORTF.OUT = (value)
    #define ADBUS_read()         PORTF.IN
    
    // swap bit order of the address/data bus
    #define SWAP_ADBUS           1
    
    // Control signals to the ethernet chip
    #define ALE(value)           writeBit(PORTE.OUT,7,value)
    #define nWR(value)           writeBit(PORTE.OUT,5,value)
    #define nRD(value)           writeBit(PORTE.OUT,6,value)
    #define nCSETH(value)        writeBit(PORTE.OUT,4,value)
    #define nRSTETH(value)       writeBit(PORTA.OUT,6,value)
       
    // Status LED
    #define STATUS_LED_on()      PORTA.OUTCLR=(1<<3)
    #define STATUS_LED_off()     PORTA.OUTSET=(1<<3)
          
    // Initialisation of the IO pins for ethernet, status LED's and Reset jumper.
    #define init_io_pins()       { \
                                   PORTA.PIN3CTRL = PORT_OPC_WIREDAND_gc; \
                                   PORTA.PIN4CTRL = PORT_OPC_WIREDAND_gc; \
                                   PORTA.PIN6CTRL = PORT_OPC_WIREDAND_gc; \
                                   PORTA.OUTSET = (1<<3)+(1<<4)+(1<<6); \
                                   PORTA.DIRSET = (1<<3)+(1<<4)+(1<<6); \
                                   PORTE.OUTSET = (1<<4)+(1<<5)+(1<<6); \
                                   PORTE.DIRSET = (1<<4)+(1<<5)+(1<<6)+(1<<7); \
                                 }   
                                 
    // The serial port to be used 
    #ifdef SERIAL
        #define SERIALPORT USARTE0
        #define SERIALPORT_RXC_vect USARTE0_RXC_vect
        #define init_serial_pins()       { \
                                            PORTE.OUTSET = (1<<3); \
                                            PORTE.DIRSET = (1<<3); \
                                          }  
    #endif
                                 
    // Configuration reset jumper, see photo reset_CrumbX1-NET.png
    #define RST_JUMPER_init()           { }
    #define RST_JUMPER_write(value)     writeBit(PORTA.OUT,4,value)
    #define RST_JUMPER_read             readBit(PORTA.IN,3)
    #define RST_JUMPER_exit()           PORTA.OUTSET = (1<<4)

    // Only 2 ADC inputs are externally available on the module (PA0, PA1)
    #define ADC_CHANNELS                2
    
    // Settings for the micro SD/MMC card slot, See additional settings in avrfat32/mmc_config.h
    #ifdef CARDSLOT     
    
        // Chip select line, plus use the red LED as activity indicator
        #define nCSMMC(value)           { \
                                          if (value==0) { \
                                             PORTE.OUTCLR=(1<<1); \
                                             PORTA.OUTCLR=(1<<4); \
                                          } else { \
                                             PORTE.OUTSET=(1<<1); \
                                             PORTA.OUTSET=(1<<4); \
                                          } \
                                        }

        // Port that is used for the SPI interface
        #define MMC_PORT                PORTD
        #define MMC_SPI                 SPID
        
        // set nCSMMC, SS, MOSI and SCK to output mode. MISO is an input. The red LED is output.
        #define init_mmc_pins()         { \
                                          PORTD.DIRSET=(1<<4)+(1<<5)+(1<<7); \
                                          PORTE.DIRSET=(1<<1); \
                                          PORTE.OUTSET=(1<<1); \
                                          PORTA.DIRSET=(1<<4); \
                                        }
    #endif // CARDSLOT
    
    // Settings for the Email application
    #ifdef EMAIL_APP
       #define email_PORT       PORTC.IN
    #endif  // EMAIL_APP

    // Definition of i/o port bits that are protected.
    // These pins cannot be accessed via io-commands d, p and o.
    // 0x00 means that the whole port is not protected.
    #define protect_PORTA        0x40
    #define protect_PORTB        0x00
    #define protect_PORTC        0x00
    #ifdef CARDSLOT
         // When the card slot is used, PD4 must be a protected output.
        #define protect_PORTD    0xF0        
    #else
        #define protect_PORTD    0x00 
    #endif
    #define protect_PORTE        0xFF
    #define protect_PORTF        0xFF    
#endif


// CrumbX1-NET version 1.0 or 1.1 from www.chip45.com
//
// with 1-4 shift registers 74HC165 for 8-32 additional inputs
// PD0 = Data in     (Pin 9 Q7)
// PD1 = Load pulse  (Pin 1 nPL)
// PD4 = Shift pulse (Pin 2 CP)
//
// with 1-4 shift registers 74HC595 for 8-32 additional outputs
// PD2 = Data out    (Pin 14 DS)
// PD3 = Store pulse (Pin 12 STcp)
// PD4 = Shift pulse (Pin 11 SHcp)
//
// PD4 ist used to clock all input and output shift registers.
//
// Please note that the shift register extension has not been tested with a CrumbX1 module!
// If you tested it please report the result to stefan@stefanfrings.de.

#ifdef CrumbX1_NET_XL
    // Macros for access to address/data bus
    #define ADBUS_output()       PORTF.DIR = 255
    #define ADBUS_input()        PORTF.DIR = 0 
    #define ADBUS_write(value)   PORTF.OUT = (value)
    #define ADBUS_read()         PORTF.IN
    
    // swap bit order of the address/data bus
    #define SWAP_ADBUS           1
    
    // Control signals to the ethernet chip
    #define ALE(value)           writeBit(PORTE.OUT,7,value)
    #define nWR(value)           writeBit(PORTE.OUT,5,value)
    #define nRD(value)           writeBit(PORTE.OUT,6,value)
    #define nCSETH(value)        writeBit(PORTE.OUT,4,value)
    #define nRSTETH(value)       writeBit(PORTA.OUT,6,value)
       
    // Status LED
    #define STATUS_LED_on()      PORTA.OUTCLR=(1<<3)
    #define STATUS_LED_off()     PORTA.OUTSET=(1<<3)
          
    // Initialisation of the IO pins for ethernet, status LED's, Reset jumper and shift 
    #define init_io_pins()       { \
                                   PORTA.PIN3CTRL = PORT_OPC_WIREDAND_gc; \
                                   PORTA.PIN4CTRL = PORT_OPC_WIREDAND_gc; \
                                   PORTA.PIN6CTRL = PORT_OPC_WIREDAND_gc; \
                                   PORTA.OUTSET = (1<<3)+(1<<4)+(1<<6); \
                                   PORTA.DIRSET = (1<<3)+(1<<4)+(1<<6); \
                                   PORTA.DIRSET = (1<<1)+(1<<2)+(1<<3)+(1<<4); \
                                   PORTE.OUTSET = (1<<4)+(1<<5)+(1<<6); \
                                   PORTE.DIRSET = (1<<4)+(1<<5)+(1<<6)+(1<<7); \
                                 }   
                                 
    // The serial port to be used 
    #ifdef SERIAL
        #define SERIALPORT USARTE0
        #define SERIALPORT_RXC_vect USARTE0_RXC_vect
        #define init_serial_pins()       { \
                                            PORTE.OUTSET = (1<<3); \
                                            PORTE.DIRSET = (1<<3); \
                                          }  
    #endif
                                 
    // Configuration reset jumper, see photo reset_CrumbX1-NET.png
    #define RST_JUMPER_init()           { }
    #define RST_JUMPER_write(value)     writeBit(PORTA.OUT,4,value)
    #define RST_JUMPER_read             readBit(PORTA.IN,3)
    #define RST_JUMPER_exit()           PORTA.OUTSET = (1<<4)

    // Only 2 ADC inputs are externally available on the module (PA0, PA1)
    #define ADC_CHANNELS                2
    
    // Settings for the micro SD/MMC card slot, See additional settings in avrfat32/mmc_config.h
    #ifdef CARDSLOT     
    
        // Chip select line, plus use the red LED as activity indicator
        #define nCSMMC(value)           { \
                                          if (value==0) { \
                                             PORTE.OUTCLR=(1<<1); \
                                             PORTA.OUTCLR=(1<<4); \
                                          } else { \
                                             PORTE.OUTSET=(1<<1); \
                                             PORTA.OUTSET=(1<<4); \
                                          } \
                                        }

        // Port that is used for the SPI interface
        #define MMC_PORT                PORTD
        #define MMC_SPI                 SPID
        
        // set nCSMMC, SS, MOSI and SCK to output mode. MISO is an input. The red LED is output.
        #define init_mmc_pins()         { \
                                          PORTD.DIRSET=(1<<4)+(1<<5)+(1<<7); \
                                          PORTE.DIRSET=(1<<1); \
                                          PORTE.OUTSET=(1<<1); \
                                          PORTA.DIRSET=(1<<4); \
                                        }
    #endif // CARDSLOT
    
    // Settings for the Email application
    #ifdef EMAIL_APP
       #define email_PORT       PORTC.IN
    #endif  // EMAIL_APP

    // Definition of i/o port bits that are protected.
    // These pins cannot be accessed via io-commands d, p and o.
    // 0x00 means that the whole port is not protected.
    #define protect_PORTA        0x40
    #define protect_PORTB        0x00
    #define protect_PORTC        0x00
    #ifdef CARDSLOT
        #define protect_PORTD    0xFE     
    #else
        #define protect_PORTD    0x1E 
    #endif
    #define protect_PORTE        0xFF
    #define protect_PORTF        0xFF    

    // Number of shift register inputs and outputs
    #define XL_INPUTS            24
    #define XL_OUTPUTS           24
    
    // Macros for additional inputs through shift registers (Port X)
    // Some commands are repeated several times to slow down the serial interface a little
    #define SHIFTin              { \
                                   PORTD.OUTSET=(1<<4); \
                                   PORTD.OUTSET=(1<<4); \
                                   PORTD.OUTSET=(1<<4); \
                                   PORTD.OUTCLR=(1<<4); \
                                   PORTD.OUTCLR=(1<<4); \
                                   PORTD.OUTCLR=(1<<4); \
                                 }
    #define DATAin               readBit(PORTD.IN,0)
    #define LOADin               { \
                                   PORTD.OUTCLR=(1<<0); \
                                   PORTD.OUTCLR=(1<<0); \
                                   PORTD.OUTCLR=(1<<0); \
                                   PORTD.OUTSET=(1<<0); \
                                   PORTD.OUTSET=(1<<0); \
                                   PORTD.OUTSET=(1<<0); \
                                   PORTD.OUTSET=(1<<0); \
                                 } 
    
    // Macros for additional outputs through shift registers (Port Y)
    // Some commands are repeated several times to slow down the serial interface a little
    #define STOREout             { \
                                   PORTD.OUTSET=(1<<3); \
                                   PORTD.OUTSET=(1<<3); \
                                   PORTD.OUTSET=(1<<3); \
                                   PORTD.OUTCLR=(1<<3); \
                                   PORTD.OUTCLR=(1<<3); \
                                   PORTD.OUTCLR=(1<<3); \
                                 }
    #define SHIFTout             { \
                                   PORTD.OUTSET=(1<<4); \
                                   PORTD.OUTSET=(1<<4); \
                                   PORTD.OUTSET=(1<<4); \
                                   PORTD.OUTCLR=(1<<4); \
                                   PORTD.OUTCLR=(1<<4); \
                                   PORTD.OUTCLR=(1<<4); \
                                 }
    #define DATAout(value)       writeBit(PORTD.OUT,2,value)

#endif


#endif // _HW_LAYOUT_H__
