#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "driver/CP2200.h"
#include "uip/uip.h"
#include "uip/uip_arp.h"
#include "httpd/httpd.h"
#include "socketd/socketd.h"
#include "uip/timer.h"
#include "io-commands.h"
#include "driver/ADC.h"
#include "hw-layout.h"

#ifdef EMAIL_APP
    #include "email_app/email_app.h"
#endif //EMAIL_APP

#ifdef SERIAL
    #include "seriald/seriald.h"
#endif //SERIAL

#ifdef SHIFTout
    #include "driver/shiftregisters.h"
#endif

#ifdef CARDSLOT
    #include "avrfat32/file.h"
    #include "avrfat32/fat.h"
    #include "avrfat32/mmc.h"
    
    volatile uint8_t card_present=0; // set to 1 during start-up if the SD card is accessible
    volatile uint8_t card_busy=0;    // Must checked and set whenever the SDMMC card is used to 
                                     // prevent that multiple threads use the avrfat32 library simultaneously.
    
    void initCardSlot(void) {
        init_mmc_pins();   
        // try to initialize the SD/MMC card up to 5 times 
        for (int i=0; i<5; i++) {   
            if (mmc_init()==1) {
                card_present=1;
                break;
            }
        }
        // read file allocation table
        if (card_present) {
            if(fat_loadFatData()==0) {
                card_present=0;
            }
        }
        // For debugging: list content of root directory
        if (card_present) {
          ffls(stdout);
        }
        
    }    
#endif

// ethernet buffer
#define BUF ((struct uip_eth_hdr *)&uip_buf[0])

// Timers used in the uip main loop
struct timer periodic_timer, arp_timer;


// Set hardcoded default IP settings
void set_defaults(void) {
    config.dhcpenable=1;
    uip_ipaddr(config.ipaddr, 192,168,42,222);
    uip_ipaddr(config.netmask, 255,255,255,0);
    uip_ipaddr(config.gateway, 192,168,42,1);
    // clear password
    config.authentication[0]=0; 
    writeFlashConfig();
};

void IP_config(void) {
#if UIP_UDP
    // if DHCP is enabled, then initialize via DHCP, else use static settings
    if (config.dhcpenable) {
        dhcpc_init();
    }
    else {
        uip_sethostaddr(config.ipaddr);
        uip_setnetmask(config.netmask);
        uip_setdraddr(config.gateway);
    }
#else // UIP_UDP
    // configure the interface with static settings
    uip_sethostaddr(config.ipaddr);
    uip_setnetmask(config.netmask);
    uip_setdraddr(config.gateway);
#endif // UIP_UDP
}


// Check if the config-reset jumper is connected
// and if yes, then reset the configuration to default settings.
uint8_t reset_jumper_connected(void) {
    RST_JUMPER_init();
    // toggle the output pin several times and check if the input pin follows
    uint8_t found=1;
    for (uint8_t i=0; i<5; i++) {
        RST_JUMPER_write(1);
        _delay_ms(1);
        if (RST_JUMPER_read==0)
            found=0;
        RST_JUMPER_write(0);
        _delay_ms(1);
        if (RST_JUMPER_read!=0)
            found=0;
    }
    if (found) {   
        // wait until the jumper gets removed
        while (RST_JUMPER_read==0) {};
    }
    RST_JUMPER_exit();
    return found;
}

// main loop, dispatches network and timer events
void uip_main_loop(void) {
    uip_len = network_device_read();
    if (uip_len > 0) {
        if (BUF->type == htons(UIP_ETHTYPE_IP)) {
            uip_arp_ipin();
            uip_input();
            // If the above function invocation resulted in data that
            // should be sent out on the network, the global variable
            // uip_len is set to a value > 0. 
            if (uip_len > 0) {
                uip_arp_out();
                network_device_send();
            }
        } else if (BUF->type == htons(UIP_ETHTYPE_ARP)) {
            uip_arp_arpin();
            // If the above function invocation resulted in data that
            // should be sent out on the network, the global variable
            // uip_len is set to a value > 0. 
            if (uip_len > 0) {
                network_device_send();
            }
        }
    }
    else if (timer_expired(&periodic_timer)) {
        timer_reset(&periodic_timer);

        for (int i = 0; i < UIP_CONNS; i++) {
            uip_periodic(i);
            // If the above function invocation resulted in data that
            // should be sent out on the network, the global variable
            // uip_len is set to a value > 0. 
            if (uip_len > 0) {
                uip_arp_out();
                network_device_send();
            }
        }

    #if UIP_UDP
        for (int i = 0; i < UIP_UDP_CONNS; i++) {
            uip_udp_periodic(i);
            // If the above function invocation resulted in data that
            // should be sent out on the network, the global variable
            // uip_len is set to a value > 0.
            if (uip_len > 0) {
                uip_arp_out();
                network_device_send();
            }
        }
    #endif // UIP_UDP

        // Call the ARP timer function every 10 seconds. 
        if (timer_expired(&arp_timer)) {
            timer_reset(&arp_timer);
            uip_arp_timer();
        }
    }
}

int main(void) {   
    
    // Set clock source to 32Mhz R/C Oscillator, only on ATxmega.
    // Atmega targets are configured via fuses.
    #ifdef OSC // ATxmega
        OSC.CTRL |= 0x02;
        while(!(OSC.STATUS & OSC_RC32MRDY_bm));
        CCP = CCP_IOREG_gc;
        CLK.CTRL = 0x01;
    #endif
    
    // Initialize direction and level of I/O pins for ethernet, status LED and reset jumper
    init_io_pins();  
             
    // Flash the status LED two times to indicate that the AVR controller is living
    STATUS_LED_on();
    _delay_ms(20);
    STATUS_LED_off();
    _delay_ms(100);
    STATUS_LED_on();
    _delay_ms(20);
    STATUS_LED_off(); 
    
    // Start system clock (timer)
    clock_init();
    
     #ifdef SERIAL
        // initialize the serial port
        #ifdef init_serial_pins
            init_serial_pins();
        #endif
        seriald_init();
        // From now, funtions like puts() and printf() can be used.
        // But serial input is processed in the main loop further below.
    #endif

    #ifdef SHIFTout
        // set all shift register outputs to LOW.
        write_XL_outputs(0);
    #endif
        
    // Reset and initialize the ethernet controller
    // The extra delay is necessary, otherwise the ethernet controllers fails sometimes.
    _delay_ms(200);
    eth_RESET();
    // Read configuration from flash (of the ethernet controller)
    readFlashConfig();
    // Initialize physical connection and MAC
    init_PHY();
    init_MAC(); 
             
   
    // Replace IP address and password by hardcoded defaults if the flash is unprogrammed (all bits 1)
    if (uip_ipaddr_cmp(config.ipaddr, all_ones_addr) &&
        uip_ipaddr_cmp(config.netmask, all_ones_addr) &&
        uip_ipaddr_cmp(config.gateway, all_ones_addr)) {
        set_defaults();
    }    
    
    // Reset configuration if the config reset jumper is connected.
    if (reset_jumper_connected()) {
        set_defaults(); 
    }

    #if ADC_CHANNELS > 0
        // initialize ADC converter
        initADC();
    #endif  
                  
    #ifdef CARDSLOT
        initCardSlot();
    #endif
    
    // Initialize uIP protocol
    uip_arp_init();
    uip_init();
    
    // Configure the IP-Adrdess
    IP_config();
        
#ifdef EMAIL_APP
    // Initialize the email application
    email_app_init();
#endif 

    // Initialize the socket daemon (telnet)
    inetd_init();
    
    // Start UIP timers
    timer_set(&periodic_timer, CLOCK_SECOND); 
    timer_set(&arp_timer, CLOCK_SECOND*10);  

    // main loop    
    while (1) {
        // Do whatever UIP need to do in the main loop
        uip_main_loop();
        
        // Execute the email application, if enabled
        #ifdef EMAIL_APP
            email_app_main_loop();
        #endif
        
        // Process serial input, if enabled
        #ifdef SERIAL
            seriald_main_loop();
        #endif
    }

    return 0;
}

